import 'dart:async';

import 'package:ble_connectivity/ble_connectivity_target_platform.dart';
import 'package:ble_connectivity/model/fhp_ble_charatereristic.dart';
import 'package:ble_connectivity/model/fhp_ble_device.dart';
import 'package:ble_connectivity/model/scan_properties.dart';
import 'package:ble_connectivity/model/sensor_data_result.dart';
import 'package:flutter_blue/flutter_blue.dart';

import 'fhp_mobile_ble_device.dart';
import 'mobile_utils.dart';

class BleConnectivityMobile implements BleConnectivityTargetPlatform {
  @override
  Function(List<FhpBleDevice> sensors) scanResultsCallback;
  FlutterBlue _flutterBlue;
  bool isScanning = false;

  @override
  ScanProperties scanProperties;

  StreamSubscription _iotipsSubscription;
  StreamSubscription _isScanningSubscription;

  /// Filter fields
  String _sensorPrefixFilter;

  List<StreamSubscription> characteristicsListSubs = [];

  StreamController<ScanEndReason> scanStreamController;

  @override
  void initScan(ScanProperties scanProperties) {
    this.scanProperties = scanProperties;
    _flutterBlue = FlutterBlue.instance;

    _flutterBlue.state.listen((event) async {
      print('ble status $event');
      if (event == BluetoothState.off)
        scanStreamController?.add(ScanEndReason.bleOff);
    });
  }

  @override
  void disposeListeners() {
    _flutterBlue.stopScan();
    _iotipsSubscription.cancel();
    _isScanningSubscription.cancel();
  }

  @override
  connectWithSensor(
      FhpBleDevice device, Function(BleDeviceState) connectionState) async {
    FhpMobileBleDevice libSensor = (device as FhpMobileBleDevice);
    libSensor.libDevice.state.listen((event) {
      connectionState(event.toFhpState());
    });
    await libSensor.connect();
  }

  @override
  Future<Stream<ScanEndReason>> startScan() async {
    if (await _flutterBlue.isOn) {
      scanStreamController = StreamController.broadcast();

      _sensorPrefixFilter = scanProperties?.sensorPrefixFilter;
      _flutterBlue
          .startScan(
            timeout: Duration(seconds: scanProperties?.scanDuration ?? 10),
          )
          .then((value) => scanStreamController?.add(ScanEndReason.ended));
    } else {
      scanStreamController?.add(ScanEndReason.bleOff);
    }
    return scanStreamController.stream;
  }

  @override
  stopScan() {
    scanStreamController?.add(ScanEndReason.canceled);
    _flutterBlue.stopScan();
    scanStreamController.close();
    scanStreamController = null;
  }

  @override
  void scanResults(Function(List<FhpBleDevice> sensors) result) {
    scanResultsCallback = result;

    _iotipsSubscription = _flutterBlue.scanResults.listen((results) {
      var iotips = results
          .where((it) => it.device.name.startsWith(_sensorPrefixFilter))
          .map((e) => FhpMobileBleDevice(
                e.device,
                connectionStatus: BleDeviceState.disconnected,
                rssi: e.rssi,
                name: e.device.name,
              ))
          .toList();
      scanResultsCallback(iotips);
      // do something with scan results
    });
    _isScanningSubscription = _flutterBlue.isScanning.listen((event) {
      isScanning = event;
    });
  }

  @override
  disconnectWithSensor(FhpBleDevice device) {
    (device as FhpMobileBleDevice).disconnect();
  }

  Future<List<Stream<SensorDataResult>>> readSensorCharacteristics(
    FhpBleDevice device,
    Map<String, List<FhpBleCharacteristic>> characteristics,
  ) async {
    List<Stream<SensorDataResult>> characteristicsStreams = [];

    List<BluetoothService> services =
        await (device as FhpMobileBleDevice).libDevice.discoverServices();

    for (var key in characteristics.keys) {
      var value = characteristics[key];
      var service = services.firstWhere((it) => it.uuid.toString() == key);
      for (FhpBleCharacteristic characteristic in value) {
        var bleCharacteristic = service.characteristics
            .firstWhere((it) => it.uuid.toString() == characteristic.uuid);

        if (characteristic.hasConfiguration) {
          await bleCharacteristic.write(characteristic.configValue);
        }

        if (characteristic.receiveNotifications) {
          await bleCharacteristic.setNotifyValue(true);
          characteristicsStreams.add(bleCharacteristic.value
              .transform(_streamTransformer(characteristic.uuid, device)));
        }
      }
    }

    return characteristicsStreams;
  }

  StreamTransformer<List<int>, SensorDataResult> _streamTransformer(
    String uuid,
    FhpBleDevice device,
  ) =>
      StreamTransformer.fromHandlers(handleData: (data, sink) {
        sink.add(
          SensorDataResult(
              value: data, characteristicUuid: uuid, device: device),
        );
      });

  @override
  Future<List<int>> readCharacteristic(FhpBleDevice device,
      {String serviceUuid, String characteristicUuid}) async {
    List<BluetoothService> services =
        await (device as FhpMobileBleDevice).libDevice.discoverServices();
    var service =
        services.firstWhere((it) => it.uuid.toString() == serviceUuid);
    var bleCharacteristic = service.characteristics
        .firstWhere((it) => it.uuid.toString() == characteristicUuid);
    return bleCharacteristic.read();
  }
}
