// Helpers for working with Pandlets
const ACCELEROMETER_UNIT_FACTOR = (4 / 32768) * 9.81;
const GYROSCOPE_UNIT_FACTOR = (2000 / 32768) * (Math.PI / 180);

// UUIDs of GATT services and charasteristics


const CONNECTION_STATE = {
    CONNECTED: 0,
    CONNECTING: 1,
    DISCONNECTED: 2
}

var errorTimeout;

class BleDevice {
    constructor() {
        this.mServer = null;
        this.position = null;
        this.bluetoothDevice = null;
        this.connectionState = CONNECTION_STATE.DISCONNECTED;
        this.connectionEventCallback = null;
        this.connectionFallbackFail = null;
        this.sensorEventCalback = null;
    }

    async clickConnect(sensorNamePrefix, scanDuration, characteristics, connectionEventCallback,
        sensorEventCalback, connectionFallbackFail) {
        var thisBleDevice = this;

        try {
            if (this.connectionState != CONNECTION_STATE.DISCONNECTED) {
                throw "Invalid state";
            }

            //start timeout error
            //errorTimeout = setTimeout(onConnectionFail, scanDuration, thisBleDevice);

            this.connectionState = CONNECTION_STATE.CONNECTING;
            this.connectionEventCallback = connectionEventCallback;
            this.connectionFallbackFail = connectionFallbackFail;
            this.sensorEventCalback = sensorEventCalback;

            let options = {};
            if (sensorNamePrefix === null || sensorNamePrefix == '') {
                options.acceptAllDevices = true;
            } else {
                let filters = [];
                filters.push({
                    namePrefix: sensorNamePrefix
                });
                options.filters = filters;
                options.optionalServices = jsonToMap(characteristics).keys();
            }

            const device = await navigator.bluetooth.requestDevice(options).then(device => {
                    thisBleDevice.bluetoothDevice = device;
                    thisBleDevice.position = device.name;
                    thisBleDevice.bluetoothDevice.addEventListener('gattserverdisconnected', () => {
                        console.log('> Bluetooth Device disconnected');
                        thisBleDevice.connectionState = CONNECTION_STATE.DISCONNECTED;

                        if (thisBleDevice.connectionEventCallback != null) {
                            console.log('Lost connection to: ' + thisBleDevice.bluetoothDevice.name);
                            if (this.connectionEventCallback == null)
                                return;
                            thisBleDevice.connectionEventCallback(false);
                        }
                        thisBleDevice.connect();
                    });
                    thisBleDevice.connect();
                })
                .catch(error => {
                    console.log('Argh! ' + error);
                    thisBleDevice.connectionFallbackFail(error.name);
                });
        } catch (err) {
            console.log('_Argh! ' + error);
            onConnectionFallbackFail(thisBleDevice);
            throw err;
        }
    };

    connect() {
       // clearTimeout(errorTimeout);

        var thisBleDevice = this;
        if (this.connectionEventCallback == null)
            return;

        this.exponentialBackoff(4 /* max retries */ , 2 /* seconds delay */ ,
            function toTry() {
                console.log('Connecting to Bluetooth Device... ');
                //modal.style.display = "block";
                return thisBleDevice.bluetoothDevice.gatt.connect();
            },
            async function success(server) {
                    thisBleDevice.mServer = server;

                    thisBleDevice.connectionState = CONNECTION_STATE.CONNECTED;

                    console.log('Connected to: ' + thisBleDevice.bluetoothDevice.name);

                    thisBleDevice.connectionEventCallback(true);
                },
                function fail() {
                    console.log('Failed to reconnect.');
                    onConnectionFallbackFail(thisBleDevice);
                });
    }

    async readCharacteristics(characteristics) {
        var thisBleDevice = this;
        var map = jsonToMap(characteristics);
        console.log(map);
        for (var serviceUuid of map.keys()) {
            var serviceCharacteristics = map.get(serviceUuid);
            const service = await thisBleDevice.mServer.getPrimaryService(serviceUuid);
            for (let index2 of  serviceCharacteristics.keys()) {
                var charValues = serviceCharacteristics.get(index2);
                listenCharacteristic(charValues, service, thisBleDevice);
            }
        }
    }


    // This function keeps calling "toTry" until promise resolves or has
    // retried "max" number of times. First retry has a delay of "delay" seconds.
    // "success" is called upon success.
    exponentialBackoff(max, delay, toTry, success, fail) {
        var thisBleDevice = this;
        toTry().then(result => success(result))
            .catch(_ => {
                if (max === 0) {
                    return fail();
                }
                if (this.connectionEventCallback == null)
                    return;
                console.log('Retrying in ' + delay + 's... (' + max + ' tries left)');
                setTimeout(function() {
                    thisBleDevice.exponentialBackoff(--max, delay * 2, toTry, success, fail);
                }, delay * 1000);
            });
    }

    disconnect() {
        if (this.connectionState != CONNECTION_STATE.CONNECTED) {
            console.log('Disconnecting:' + this.position + ' Not Connected');
            return;
        }

        console.log('Disconnecting from: ' + this.bluetoothDevice.name);
        this.connectionEventCallback = null; //To prevent callback
        this.bluetoothDevice.gatt.disconnect();
        this.connectionState = CONNECTION_STATE.DISCONNECTED;

        console.log('Disconnected from: ' + this.bluetoothDevice.name);
    }

    getCurrentValue(service, characteristic) {
        var thisBleDevice = this;
        console.log('getCurrentValue ' +thisBleDevice.bluetoothDevice.name + ', service ' +service+ ', characteristic ' +characteristic);
        var server = thisBleDevice.mServer;

        server.getPrimaryService(service).then(s => {
                return s.getCharacteristic(characteristic);
            })
            .then(c => {
                console.log('Reading characteristic...');
                return c.readValue();
            })
            .then(value => {
            if (thisBleDevice.sensorEventCalback == null) {
                return;
            }
            console.log('sensorEventCalback...');
                thisBleDevice.sensorEventCalback(characteristic, new Uint8Array(value.buffer));
              //  thisBleDevice.sensorEventCalback(characteristic, value.getUint8(0));
            })
            .catch(error => {
                console.log('Argh! ' + error);
            });
    }
}

function jsonToMap(jsonString)  {
    var jsonObject = JSON.parse(jsonString);
    var dataMap = new Map(Object.entries(jsonObject));
    var resultMap = new Map();
    for (const key of dataMap.keys())  {
        var keyMap = new Map(Object.entries(dataMap.get(key)));
        resultMap.set(key, keyMap);
    }

    return resultMap;
}

function onConnectionFallbackFail(bleDevice) {
    console.log('onConnectionFallbackFail');
    bleDevice.connectionFallbackFail('fail');
    onConnectionFail(bleDevice);
};

function onConnectionFail(bleDevice) {
    bleDevice.connectionState = CONNECTION_STATE.DISCONNECTED;
    bleDevice.connectionEventCallback(false);
    //stop handler
   // clearTimeout(errorTimeout);
};


function time(text) {
    console.log('[' + new Date().toJSON().substr(11, 8) + '] ' + text);
}

async function  listenCharacteristic (charValues, service, thisBleDevice) {
    var characteristicUuid = charValues['uuid'];
    var configValue = charValues['configValue'];
    var receiveNotifications = charValues['receiveNotifications'];
    const serviceCharacteristic = await service.getCharacteristic(characteristicUuid)

    if(configValue != null){
        await serviceCharacteristic.writeValue(Uint8Array.of(configValue));
    }
    if(receiveNotifications) {
        await serviceCharacteristic.startNotifications()
        serviceCharacteristic.addEventListener('characteristicvaluechanged', (event) => {
            if (thisBleDevice.sensorEventCalback == null) {
                return;
            }

            thisBleDevice.sensorEventCalback(characteristicUuid, new Uint8Array(event.target.value.buffer));
        });
    }
}